.BaseEnv <- NULL
.NameSpaces <- NULL
make.env <- function(parent = .GlobalEnv, hashed = FALSE) {
    if (hashed) {
        env <- attach(NULL)
        detach(2)
        .Call("Namespace_reparent", env, parent, PACKAGE="namespace")
    }
    else eval(quote((function() environment())()), env=parent)
}
register.namespace <- function(ns, name = namespace.name(ns)) {
    name <- as.character(as.name(name))
    if (exists(name, env = .NameSpaces))
        warning(paste("namespace \"", name ,"\" already exists", sep=""))
    assign(name, ns, env = .NameSpaces)
}
get.namespace <- function(name) {
    name <- as.character(name)
    get(name, env=.NameSpaces)
}
namespace.exists <- function(name) {
    name <- as.character(name)
    exists(name, env=.NameSpaces)
}
as.namespace <- function(ns) {
    if (is.character(ns)) get.namespace(ns)
    else if (inherits(ns,"namespace")) ns
    else stop("bad namespace argument")
}
make.import.export.names <- function(spec) {
    old <- as.character(spec)
    new <- names(spec)
    if (is.null(new)) new <- old
    else new[new==""] <- old[new==""]
    names(old) <- new
    old
}
namespace.eval <- function(expr, ns) eval(expr, env=ns$env())
attach.namespace <- function(ns, prefix="namespace") {
    ns <- as.namespace(ns)
    exports <- namespace.exports(ns)
    nsget <- function(n) namespace.get(ns, n)
    frame <- structure(lapply(exports, nsget), names=exports)
    framename <- paste(prefix, namespace.name(ns),sep=":")
    attach(frame, name = framename)
}
namespace.file.path <- function(package, package.lib)
    file.path(package.lib, package, "NAMESPACE")
namespace.file.exists <- function(package, package.lib)
    file.exists(namespace.file.path(package, package.lib))
parse.namespace.file <- function(package, package.lib) {
    nsFile <- namespace.file.path(package, package.lib)
    if (! file.exists(nsFile)) {
        fQuote <- function(s) paste("`", s, "'", sep = "")
        stop(paste("package", fQuote(package), "has no NAMESPACE file"))
    }
    exports <- character(0)
    imports <- list()
    for (e in parse(nsFile))
        switch(as.character(e[[1]]),
               export = {
                   exp <- e[-1]
                   exp <- structure(as.character(exp), names=names(exp))
                   exports <- c(exports, exp)
               },
               import = imports <- c(imports,as.list(as.character(e[-1]))),
               import.from = {
                   imp <- e[-1]
                   ivars <- imp[-1]
                   inames <- names(ivars)
                   imp <- list(as.character(imp[1]),
                               structure(as.character(ivars), names=inames))
                   imports <- c(imports, list(imp))
               },
               stop(paste("unknown namespace directive:", deparse(e))))
    list(imports=imports, exports=exports)
}
load.namespace <- function (package, lib.loc = .lib.loc,
                            keep.source = getOption("keep.source.pkgs")) {
    package <- as.character(package)
    if (namespace.exists(package))
        get.namespace(package)
    else {
        fQuote <- function(s) paste("`", s, "'", sep = "")
        package.lib <- lib.loc[file.exists(file.path(lib.loc, package))]
        if (length(package.lib) == 0)
            stop(paste("There is no package called", fQuote(package)))
        else if (length(package.lib) > 1) {
            package.lib <- package.lib[1]
            warning(paste("Package ", fQuote(package),
                          "found more than once,\n", 
                          "using the one found in", fQuote(package.lib)))
        }

        # create namespace and process imports
        #**** FIXME: should do something to catch circular dependencies
        nsInfo <- parse.namespace.file(package, package.lib)
        ns <- make.namespace(package, register = FALSE)
        for (i in nsInfo$imports) {
            if (is.character(i))
                ns$import(load.namespace(i, lib.loc, keep.source))
            else
                ns$import.from(load.namespace(i[[1]], keep.source), i[[2]])
        }
        env <- ns$env()

        codeFile <- file.path(package.lib, package, "R", package)
        if (file.exists(codeFile)) 
            sys.source(codeFile, env, keep.source = keep.source)
        else warning(paste("Package ", fQuote(package), "contains no R code"))

        if (exists(".First.lib", envir = env, inherits = FALSE)) {
             firstlib <- get(".First.lib", envir = env, inherits = FALSE)
             run.First.lib(firstlib, env, package, package.lib)
         }
         firstlib <- getOption(".First.lib")[[package]]
         if (!is.null(firstlib))
             run.First.lib(firstlib, env, package, package.lib)

         # export, seal, and register
         # registration is done here to avoid registering a namespace
         # with errors
         ns$export(nsInfo$exports)
         ns$seal()
         register.namespace(ns)

         ns
     }                               
}
run.First.lib <- function(fun, env, package, package.lib) {
    opt <- options(standardDatabase = where)
    on.exit(options(opt))
    options(LoadingNameSpaceEnv = env)
    tt <- try(firstlib(which.lib.loc, package))
    if (inherits(tt, "try-error"))
        stop(".First.lib failed")
}
make.namespace <- function(name = NULL, register = TRUE) {
    exports <- character()
    impenv <- make.env(.BaseEnv, hashed=TRUE)
    env <- make.env(impenv, hashed=TRUE)
    sealed <- FALSE
    imports <- list("base"=TRUE)
    self <- list(import = function(...) {
                     if (sealed) stop("namespace is sealed")
                     for (ns in list(...)) {
                         ns <- as.namespace(ns)
                         self$import.from(ns)
                     }
                 },
                 import.from = function(ns, vars) {
                     #**** FIXME: need to worry about uniqueness of names?
                     if (sealed) stop("cannot import into a sealed namespace")
                     ns <- as.namespace(ns)
                     if (missing(vars)) impvars <- ns$exports()
                     else impvars <- vars
                     impvars <- make.import.export.names(impvars)
                     impnames <- names(impvars)
                     for (i in seq(along = impvars))
                         assign(impnames[i], ns$get(impvars[i]), env=impenv)
                     if (missing(vars)) imp <- TRUE
                     else imp <- impvars
                     imp <- structure(list(imp), names = ns$name())
                     imports <<- c(imports, imp)
                 },
                 imports = function(ns) imports,
                 env = function() {
                     if (sealed) stop("cannot get environment of a sealed namespace")
                     env
                 },
                 export = function(new) {
                     #**** FIXME: need to worry about uniqueness of names?
                     if (sealed) stop("cannot add to exports of a sealed namespace")
                     new <- make.import.export.names(new)
                     undef <- new[! sapply(new, exists, env=env)]
                     if (length(undef) != 0) {
                         undef <- do.call("paste", as.list(c(undef, sep=", ")))
                         stop(paste("undefined exports:", undef))
                     }
                     exports <<- c(new,exports)
                 },
                 exports = function() unique(names(exports)),
                 seal = function() {
                     if (sealed) stop("already sealed")
                     sealed <<- TRUE
                 },
                 name = function() name,
                 get = function(name) {
                     if (! sealed) stop("cannot get values from unsealed namespace")
                     intname <- exports[name]
                     if (is.na(intname))
                         stop(paste(name, "is not an exported variable"))
                     get(intname, env=env)
                 })
    class(self) <- "namespace"
    if (register && ! is.null(name)) {
        name <- as.character(as.name(name))
        register.namespace(self, name)
    }
    self
}
print.namespace <- function(ns) {
    name <- ns$name()
    if (is.null(name)) cat("<namespace>\n")
    else cat("<namespace \"", name, "\">\n", sep="")
}    
namespace.import <- function(ns, ins)
    as.namespace(ns)$import(ins)
namespace.import.from <- function(ns, ins, vars)
    as.namespace(ns)$import.from(ins, vars)
namespace.imports <- function(ns)
    as.namespace(ns)$imports()
namespace.export <- function(ns, vars)
    as.namespace(ns)$export(vars)
namespace.exports <- function(ns)
    as.namespace(ns)$exports()
namespace.env <- function(ns) as.namespace(ns)$env()
namespace.get <- function(ns, var) as.namespace(ns)$get(var)
namespace.seal <- function(ns) as.namespace(ns)$seal()
namespace.name <- function(ns) as.namespace(ns)$name()
namespace.library <-
function (package, help, lib.loc = .lib.loc, character.only = FALSE, 
    logical.return = FALSE, warn.conflicts = package != "MASS", 
    keep.source = getOption("keep.source.pkgs")) 
{
    fQuote <- function(s) paste("`", s, "'", sep = "")
    if (!missing(package)) {
        if (!character.only) 
            package <- as.character(substitute(package))
        pkgname <- paste("package", package, sep = ":")
        if (is.na(match(pkgname, search()))) {
            which.lib.loc <- lib.loc[file.exists(file.path(lib.loc, 
                package))]
            if (length(which.lib.loc) == 0) {
                txt <- paste("There is no package called", fQuote(package))
                if (logical.return) {
                  warning(txt)
                  return(FALSE)
                }
                else stop(txt)
            }
            if (length(which.lib.loc) > 1) {
                which.lib.loc <- which.lib.loc[1]
                warning(paste("Package ", fQuote(package), "found more than once,\n", 
                  "using the one found in", fQuote(which.lib.loc)))
            }
            if (namespace.file.exists(package, which.lib.loc)) {
                tt <- try({
                    ns <- load.namespace(package, which.lib.loc)
                    attach.namespace(ns, prefix="package")
                })
               if (inherits(tt, "try-error")) 
                   if (logical.return) 
                       return(FALSE)
                   else stop("package/nameslace load failed")
               else
                   if (logical.return)
                       return(TRUE)
                   else
                       return(invisible(.packages()))
            }
            codeFile <- file.path(which.lib.loc, package, "R", 
                package)
            env <- attach(NULL, name = pkgname)
            on.exit(detach(2))
            attr(env, "path") <- file.path(which.lib.loc, package)
            if (file.exists(codeFile)) 
                sys.source(codeFile, env, keep.source = keep.source)
            else warning(paste("Package ", fQuote(package), "contains no R code"))
            .Internal(lib.fixup(env, .GlobalEnv))
            if (exists(".First.lib", envir = env, inherits = FALSE)) {
                firstlib <- get(".First.lib", envir = env, inherits = FALSE)
                tt <- try(firstlib(which.lib.loc, package))
                if (inherits(tt, "try-error")) 
                  if (logical.return) 
                    return(FALSE)
                  else stop(".First.lib failed")
            }
            if (!is.null(firstlib <- getOption(".First.lib")[[package]])) {
                tt <- try(firstlib(which.lib.loc, package))
                if (inherits(tt, "try-error")) 
                  if (logical.return) 
                    return(FALSE)
                  else stop(".First.lib failed")
            }
            if (warn.conflicts && !exists(".conflicts.OK", envir = env, 
                inherits = FALSE)) {
                dont.mind <- c("last.dump", "last.warning", ".Last.value", 
                  ".Random.seed")
                lib.pos <- 2
                ob <- objects(lib.pos)
                fst <- TRUE
                ipos <- seq(along = sp <- search())[-c(lib.pos, 
                  match("Autoloads", sp))]
                for (i in ipos) {
                  obj.same <- match(objects(i), ob, nomatch = 0)
                  if (any(obj.same > 0) && length(same <- (obs <- ob[obj.same])[!obs %in% 
                    dont.mind])) {
                    if (fst) {
                      fst <- FALSE
                      cat("\nAttaching Package \"", pkgname, 
                        "\":\n\n", sep = "")
                    }
                    cat("\n\tThe following object(s) are masked", 
                      if (i < lib.pos) 
                        "_by_"
                      else "from", sp[i], ":\n\n\t", same, "\n\n")
                  }
                }
            }
            on.exit()
        }
        else {
            if (getOption("verbose")) 
                warning(paste("Package", pkgname, "already present in search()"))
        }
    }
    else if (!missing(help)) {
        if (!character.only) 
            help <- as.character(substitute(help))
        help <- help[1]
        which.lib.loc <- lib.loc[file.exists(file.path(lib.loc, 
            help))]
        if (length(which.lib.loc) == 0) 
            stop(paste("No documentation for package", fQuote(help)))
        if (length(which.lib.loc) > 1) {
            which.lib.loc <- which.lib.loc[1]
            warning(paste("Package ", fQuote(help), "found more than once,\n", 
                "using the one found in", fQuote(which.lib.loc)))
        }
        outFile <- tempfile("Rlibrary")
        docFiles <- file.path(which.lib.loc, help, c("TITLE", 
            "DESCRIPTION", "INDEX"))
        headers <- c("", "Description:\n\n", "Index:\n\n")
        footers <- c("\n", "\n", "")
        for (i in which(file.exists(docFiles))) {
            cat(headers[i], file = outFile, append = TRUE)
            file.append(outFile, docFiles[i])
            cat(footers[i], file = outFile, append = TRUE)
        }
        file.show(outFile, delete.file = TRUE, title = paste("Documentation for package", 
            fQuote(help)))
    }
    else {
        libfil <- tempfile("R.")
        avail <- NULL
        for (lib in lib.loc) {
            cat("\nPackages in library `", lib, "':\n\n", sep = "", 
                file = libfil, append = TRUE)
            if (file.exists(libind <- file.path(lib, "LibIndex"))) {
                file.append(libfil, libind)
                a <- NULL
            }
            else {
                a <- .packages(all.available = TRUE, lib.loc = lib)
                for (i in sort(a)) {
                  title <- file.path(lib, i, "TITLE")
                  if (file.exists(title)) 
                    file.append(libfil, title)
                  else cat(i, "\n", file = libfil, append = TRUE)
                }
            }
            avail <- c(avail, a)
        }
        file.show(libfil, delete.file = TRUE, title = "R packages available")
        return(invisible(avail))
    }
    if (logical.return) 
        TRUE
    else invisible(.packages())
}
library <- namespace.library
.First.lib <- function(lib, pkg) {
    library.dynam(pkg, pkg, lib)
    .BaseEnv <<- make.env(.GlobalEnv, hashed=TRUE)
    bpos <- match("package:base", search())
    exclude <- c(".Options", ".Random.seed")  #**** anything else???
    for (name in ls(bpos))
        if (! match(name, exclude, 0))
            assign(name, get(name, bpos), env = .BaseEnv)
    .NameSpaces <<- make.env(NULL)
}
